# Metaphase links
FRAMES_AROUND_METAPHASE = (
    10  # look for metaphase candidate spots in this range
)

INTERPHASE_INDEX = 0  # interphase index for CNN classification model
METAPHASE_INDEX = 1  # metaphase index for CNN classification model

# Mid-body tracking
CYTOKINESIS_DURATION = (
    20  # number of frames to look for mid-body in between cells
)

# Minimum distance to border to consider is_near_border = true
MINIMUM_DISTANCE_TO_BORDER = 20
